/**
 * e3db Fast database library for Grails
 * Copyright (C) 2009-2010 Collegeman.net, LLC
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

import grails.test.*
import net.collegeman.grails.e3db.*
import org.springframework.jdbc.datasource.*

class E3dbTests extends GrailsUnitTestCase {
    
	private static String dbVersion = '5.0.51b'
	
	static {
		System.setProperty("net.collegeman.grails.e3db.printSql", "true");
		
		DB.setDefaultDataSource(new SimpleDriverDataSource(
			new com.mysql.jdbc.Driver(),
			"jdbc:mysql://localhost/e3db",
			"root",
			"password"
		))
	}

	protected void setUp() {
		DB.withTransaction {
			DB.exec("DROP TABLE IF EXISTS test");
			DB.exec("CREATE TABLE test (name VARCHAR(128), color VARCHAR(10), featured TINYINT(1))")
			DB.batch("INSERT INTO test (name, color, featured) VALUES (?, ?, ?)", [
				['Aaron', 'red', 1],
				['Karla', 'blue', 1],
				['Izzy', 'green', 0],
				['Maria', 'black', 1]
			]);
		}
	}

    protected void tearDown() {
		DB.exec("DROP TABLE IF EXISTS test");
    }

    void testGetSql() {
		assert dbVersion == DB.get("SELECT @@VERSION")
		assert 'foo' != DB.get("SELECT @@VERSION")
    }

	void testGetClosure() {
		assert dbVersion == DB.get {
			"SELECT @@VERSION"
		}
		
		assert dbVersion == DB.get {
			sql("SELECT @@VERSION")
		}
		
		assert dbVersion == DB.get {
			sql("SELECT")
			sql("@@VERSION")
		}		
	}
	
	void testGetInt() {
		assert 4 == DB.get("SELECT COUNT(*) FROM test")
		assert 4 == DB.getInt("SELECT COUNT(*) FROM test")
		assert new Integer(4) == DB.getInt("SELECT COUNT(*) FROM test")
		assert !new Long(4).equals(DB.getInt("SELECT COUNT(*) FROM test"))
	}
	
	void testGetLong() {
		assert 4 == DB.get("SELECT COUNT(*) FROM test")
		assert 4 == DB.getLong("SELECT COUNT(*) FROM test")
		assert new Long(4) == DB.getLong("SELECT COUNT(*) FROM test")
		assert !new Integer(4).equals(DB.getLong("SELECT COUNT(*) FROM test"))
	}
	
	void testRows() {
		def tests = [ 'Aaron', 'Karla', 'Izzy', 'Maria' ]
		
		DB.rows("SELECT * FROM test").eachWithIndex { test, i ->
			assert tests[i] == test.name
		}
		
		DB.rows("SELECT * FROM test").eachWithIndex { test, i ->
			assert tests[i] != test.color
		}
	}
	
	static def params = [ go: true ]
	
	void testDelegation() {
		assert dbVersion == DB.get {
			if (params.go)
				sql("SELECT @@VERSION")
			else
				sql("SELECT 'foo'")
		}
	}
	
	void testInjunctions() {
		assert 2 == DB.get {
			sql("SELECT count(name) FROM test WHERE")
			or {
				sql("color = ?", 'red')
				sql("color = ?", 'blue')
			}
		}
		
		assert 1 == DB.get {
			sql("SELECT count(name) FROM test WHERE")
			or {
				sql("color = ?", 'red')
				sql("name = ?", 'Aaron')
			}
		}
	}
	
	void testWhere() {
		assert 1 == DB.get {
			sql("SELECT COUNT(*) FROM test")
			where {
				sql('color = ?', 'red')
			}
		}
		
		assert 2 == DB.get {
			sql("SELECT COUNT(*) FROM test")
			where {
				or {
					['red', 'blue'].each {
						sql('color = ?', it)
					}
				}
			}
		}
	}
	
	void testExists() {
		assert 1 == DB.get {
			sql("SELECT")
			exists {
				sql("SELECT name FROM test WHERE color = ?", 'red')
			}
		}
	}
	
	void testNamedParameters() {
		assert 'Aaron' == DB.get {
			sql("SELECT name FROM test WHERE color = :color", [color: 'red'])
		}
		
		assert 2 == DB.get {
			sql("SELECT COUNT(*) FROM test WHERE color IN (:color)", [color: ['red', 'blue']])
		}
	}
	
	void testBuffer() {
		assert "SELECT * FROM test WHERE color = ?" == DB.buffer {
			sql("SELECT * FROM test WHERE color = ?", 'red')
		}.getSql();
	}
	
	void testJSONCompiler() {
		def compiler = DB.JSON.createCompiler()
		
		assert "3.1415" == compiler.compile(3.1415)
		
		assert '"foo"' == compiler.compile('foo')
		
		assert '[1,3.1415,"foo"]' == compiler.compile([1, 3.1415, 'foo'])
		
		assert "true" == compiler.compile(true)
		
		assert "false" == compiler.compile(false)
		
		assert "null" == compiler.compile(null)
		
	 	assert '{"integer":1,"floating":3.1415,"string":"foo","bool":true,"nullified":null,"closure":{"integer":1,"floating":3.1415,"string":"foo"},"array":[1,3.1415,"\\"foo"]}' == compiler.compile {
			integer = 1
			floating = 3.1415
			string = 'foo'
			bool = true
			nullified = null
			closure = {
				integer = 1
				floating = 3.1415
				string = 'foo'
			}
			array = [1, 3.1415, '"foo']
		}
	}
	
	void testJSON() {
		def params = [ 
			page: 1,
			pageSize: 2,
			colors: 'red,blue' 
		]
		
		assert '{"records":[{"name":"Aaron","color":"red","featured":true},{"name":"Karla","color":"blue","featured":true}],"total":2,"success":true}' == DB.json {
			records = selectWithCount {
				// use doCount() to control query mode
				// when true, query should result in a COUNT(*) query
				// when false, query should return a paginated result set
				if (doCount()) {
				 	sql("COUNT(name)")
				}
			 	else {
					sql("name,color,featured")
				}
				sql("FROM test")
				where {
					if (params.colors) {
						sql("color IN (:colors)", [colors: Arrays.asList(params.colors.split(','))])
					}
				}
				// here, doCount() is used to control the inclusion of pagination offsets
				if (!doCount()) {
					format("LIMIT %d OFFSET %d", params.pageSize ?: 2, ((params.page ?: 1)-1) * (params.pageSize ?: 2))
				}
			}
			
			total = recordCount()
		}
		
		assert '{"records":null,"success":false,"error":{"message":"StatementCallback; bad SQL grammar [SELECT SELECT * FROM nonexistenttable]; nested exception is java.sql.SQLException: You have an error in your SQL syntax; check the manual that corresponds to your MySQL server version for the right syntax to use near \'SELECT * FROM nonexistenttable\' at line 1","cause":"java.sql.SQLException","type":"org.springframework.jdbc.BadSqlGrammarException"}}' == DB.json {
			records = select {
				sql("SELECT * FROM nonexistenttable")
			}
		}
		
		
		def exception
		try {
			DB.json().handleErrors(false).compile {
				records = select {
					sql("SELECT * FROM nonexistenttable")
				}
			}
		} catch (Exception e) {
			exception = e
		}
		assert org.springframework.jdbc.BadSqlGrammarException == exception.class
	}


}
